import { ComponentType, createElement } from 'react'
import { hydrate, render } from 'react-dom'
import { importModule, trimModuleExt } from '../core/module.ts'
import { RouteModule, Routing, RoutingOptions } from '../core/routing.ts'
import Router from './components/Router.ts'
import { loadPageDataFromTag } from './pagedata.ts'
import { createPageProps, PageRoute } from './pageprops.ts'

type BootstrapOptions = Required<RoutingOptions> & {
  sharedModules: RouteModule[],
  renderMode: 'ssr' | 'spa'
}

export default async function bootstrap(options: BootstrapOptions) {
  const { baseURL, defaultLocale, locales, routes, rewrites, sharedModules, renderMode } = options
  const { document } = window as any
  const customComponents: Record<string, { C: ComponentType, useDeno?: boolean }> = {}
  await Promise.all(sharedModules.map(async ({ url, useDeno }) => {
    const { default: C } = await importModule(baseURL, url)
    switch (trimModuleExt(url)) {
      case '/404':
        customComponents['E404'] = { C, useDeno }
        break
      case '/app':
        customComponents['App'] = { C, useDeno }
        break
    }
  }))
  const routing = new Routing({ routes, rewrites, baseURL, defaultLocale, locales })
  const [url, nestedModules] = routing.createRouter()
  const imports = nestedModules.map(async mod => {
    const { default: Component } = await importModule(baseURL, mod.url)
    return {
      url: mod.url,
      Component
    }
  })
  if (!!customComponents.App?.useDeno || nestedModules.findIndex(mod => !!mod.useDeno) > -1) {
    await loadPageDataFromTag(url)
  }
  const pageRoute: PageRoute = { ...createPageProps(await Promise.all(imports)), url }
  const routerEl = createElement(Router, { customComponents, pageRoute, routing })
  const mountPoint = document.getElementById('__aleph')

  if (renderMode === 'ssr') {
    hydrate(routerEl, mountPoint)
  } else {
    render(routerEl, mountPoint)
  }

  // remove ssr head elements, set a timmer to avoid the tab title flash
  setTimeout(() => {
    Array.from(document.head.children).forEach((el: any) => {
      const tag = el.tagName.toLowerCase()
      if (
        el.hasAttribute('ssr') &&
        tag !== 'style' &&
        !(tag === 'link' && el.getAttribute('rel') === 'stylesheet')
      ) {
        document.head.removeChild(el)
      }
    })
  }, 0)
}
